#include "Spr.h"


// Public variables
UINT8 Spr_IsMeasurementComplete;
UINT8 Spr_FlowData[SPR_OUTBUFFER_DATA_SIZE];    // Flow data buffer
UINT8 Spr_VolumeData[SPR_OUTBUFFER_DATA_SIZE];  // Volume data buffer
UINT8 Spr_Results[SPR_OUTBUFFER_RESULTS_SIZE];  // Results buffer
UINT8 Data_type;                                // Type of data to be sent [Inhalation/Exhalation/Results]
UINT8 GraphBufferIndex;

// States
static void StateInit(void);
static void StateIdle(void);
static void StateWaitForInhalation(void);
static void StateMeasuringInhalation(void);
static void StateWaitForExhalation(void);
static void StateMeasuringExhalation(void);
static void StateSendingResults(void);

// Timer events
static void TimerSampleAdcSpr_Event(void);

// Send data to PC
static void SendSprGraphDataToPc(void);   // Fills the flow and volume buffers
static void SendResultsToPc(void);        // Fills the results buffer

// Clear variables for a new measurement
static void ClearVariables(void);


// Calculate parameters
static void CalculateParameters(void);


/** Main Spr state machine states */
typedef enum
{
	STATE_INIT,
	STATE_IDLE,
	STATE_WAIT_FOR_INHALATION,
	STATE_MEASURING_INHALATION,
	STATE_WAIT_FOR_EXHALATION,
	STATE_MEASURING_EXHALATION,
        STATE_SENDING_RESULTS,
} SprStates_e;


/* Main state machine */
void (*const SprStateMachine[]) (void) =
{
	StateInit,
	StateIdle,
	StateWaitForInhalation,
	StateMeasuringInhalation,
	StateWaitForExhalation,
	StateMeasuringExhalation,
        StateSendingResults,
};

/* Private variables */
static MovingAverage_uint16_t SprPressureSignal;
static UINT8 SprActualState = STATE_INIT;
static UINT8 IsnewSampleAvailable;
static UINT8 SprActualEvent = EVENT_SPR_NONE;
static UINT16 time_counter_ms;
static UINT16 SprPressureSignalOffset = 0;

// Data
static UINT16 Volume;
static UINT16 Flow;

// Results
static UINT16 PIF;             // Peak Inspiratory Flow
static UINT16 FIVC;            // Forced Inspiratory Volume Capacity
static UINT16 PEF;             // Peak Expiratory Flow
static UINT16 FVC;             // Forced Vital Capacity
static UINT16 FEV1;            // Forced Expiratory Volume in 1 second
static UINT16 FEV1_FVC_ratio;  // Ratio of FEV1 to FVC
static UINT16 FET;             // Forced Expiratory Time


static UINT8 IgnoreInitialSamplesCounter;

// Timers
static UINT8 TimerSampleAdcSpr;


/*************************************************************/


/* States and substates */

// StateInit obtains the pressure signal offset //
static void StateInit(void)
{
static UINT16 initSamples = INIT_SAMPLES;
static UINT32 initSamplesSum = 0;

	if (IsnewSampleAvailable)
	{
		IsnewSampleAvailable = FALSE;	
		
		if (IgnoreInitialSamplesCounter > 0)
		{
			IgnoreInitialSamplesCounter--;
		}
		else
		{		
			if (initSamples > 0)
			{
				initSamples--;
				initSamplesSum += SprPressureSignal.Result;
			}
			else
			{
				SprPressureSignalOffset = (UINT16) (initSamplesSum/(INIT_SAMPLES));
				
				Spr_AbortMeasurement();
				
				initSamplesSum = 0;
			}				
		}
	}
}


// StateIdle does nothing. System is in Standby //
static void StateIdle(void)
{
  // Do nothing
}


// StateWaitForInhalation waits for the patient to start inhaling //
static void StateWaitForInhalation(void)
{
	if (IsnewSampleAvailable)
	{
		IsnewSampleAvailable = FALSE;

		if(IgnoreInitialSamplesCounter>0)
		{
			IgnoreInitialSamplesCounter--;
		}
		
		else
		{	
		  // if	signal goes below the inhalation threshold, measure inhalation		
			if (SprPressureSignal.Result < (SprPressureSignalOffset - INHALATION_DETECTION_THRESHOLD))
			{
				SprActualState = STATE_MEASURING_INHALATION;
			}
		}
	}
}


// StateMeasuringInhalation measures and obtains data during inhalation //
static void StateMeasuringInhalation(void)
{
UINT8 i;
	if (IsnewSampleAvailable)
	{
		IsnewSampleAvailable = FALSE;
		
		// if signal goes back to "zero", wait for exhalation
		if ((SprPressureSignal.Result) >= SprPressureSignalOffset)
		{
			SprActualState = STATE_WAIT_FOR_EXHALATION;
			Volume = 0;
			time_counter_ms = 0;			
			
			if (GraphBufferIndex != 0)
			{
			  // if buffers have data, send data to PC
			  for (i=GraphBufferIndex; i<SPR_OUTBUFFER_DATA_SIZE; i+=2)
			  {
			    Spr_FlowData[i] = 0;
			    Spr_FlowData[i+1] = 0;
			    Spr_VolumeData[i] = Spr_VolumeData[i-2];
			    Spr_VolumeData[i+1] = Spr_VolumeData[i-1];
			  }
			  SprActualEvent = EVENT_SPR_NEW_DATA_READY;
			  GraphBufferIndex = 0;
			}
		}
		
		else
		{
		CalculateParameters();
		
		SendSprGraphDataToPc();
		}
	}
}


// StateWaitForExhalation waits for the patient to exhale //
static void StateWaitForExhalation(void)
{
  UINT8 i;
  
	if (IsnewSampleAvailable)
	{
		IsnewSampleAvailable = FALSE;	
		
		// wait here until signal goes above the detection threshold
		if (SprPressureSignal.Result >= (SprPressureSignalOffset + INHALATION_DETECTION_THRESHOLD))
		{
		  for (i=0;i<SPR_OUTBUFFER_DATA_SIZE;i++)
		  {
		    Spr_FlowData[i] = 0;
		    Spr_VolumeData[i] = 0;
		  }
		  Data_type = EXH_DATA_TYPE;
		  SprActualEvent = EVENT_SPR_NEW_DATA_READY;
		  SprActualState = STATE_MEASURING_EXHALATION;
		}
	}	
}

	
// StateMeasuringExhalation measures and obtain data during exhalation //	
static void StateMeasuringExhalation(void)
{
UINT8 i;
	if (IsnewSampleAvailable)
	{
		IsnewSampleAvailable = FALSE;
												
		// if signal goes back to "zero", exhalation has finished
		if ((SprPressureSignal.Result) <= SprPressureSignalOffset)
		{
			if (GraphBufferIndex != 0)
			{
			  // if buffers have data, send data to PC
			  for (i=GraphBufferIndex; i<SPR_OUTBUFFER_DATA_SIZE; i+=2)
			  {
			    Spr_FlowData[i] = 0;
			    Spr_FlowData[i+1] = 0;
			    Spr_VolumeData[i] = Spr_VolumeData[i-2];
			    Spr_VolumeData[i+1] = Spr_VolumeData[i-1];
			  }
			  SprActualEvent = EVENT_SPR_NEW_DATA_READY;
			}
			Spr_IsMeasurementComplete = TRUE;
			SprActualState = STATE_SENDING_RESULTS;
		}
		
		else
		{
		  CalculateParameters();

      SendSprGraphDataToPc();
		}
	}
}


// StateSendingResults generates an event to send results to PC
static void StateSendingResults (void)
{
  UINT32 i;
  
  SendResultsToPc();
  Data_type = RESULTS;
  SprActualEvent = EVENT_SPR_NEW_DATA_READY;
  Spr_AbortMeasurement();
  for (i=0; i<140000; i++){}
}


static void SendSprGraphDataToPc(void)
{
  // Store data and send it when the buffers are full
	Spr_FlowData[GraphBufferIndex] = (UINT8)(Flow >> 8); 		        // Flow higher byte
	Spr_VolumeData[GraphBufferIndex] = (UINT8)(Volume >> 8);        // Volume higher byte
	GraphBufferIndex++;
		
	Spr_FlowData[GraphBufferIndex] = (UINT8)(Flow & 0x0FF);         // Flow lower byte
	Spr_VolumeData[GraphBufferIndex] = (UINT8)(Volume & 0x00FF);    // Volume lower byte
	GraphBufferIndex++;
		
	if (GraphBufferIndex == SPR_OUTBUFFER_DATA_SIZE)
	{
	  // Buffers are full, call the NEW_DATA_READY event
	  SprActualEvent = EVENT_SPR_NEW_DATA_READY;
	  GraphBufferIndex = 0;
	}
}


// Fill the results buffer with the final results
static void SendResultsToPc(void)
{
  UINT8 BufferIndex;
  
  BufferIndex = 0;
  Spr_Results[BufferIndex++] = (UINT8)(FIVC >> 8);
  Spr_Results[BufferIndex++] = (UINT8)(FIVC & 0x00FF);
  Spr_Results[BufferIndex++] = (UINT8)(PIF >> 8);
  Spr_Results[BufferIndex++] = (UINT8)(PIF & 0x00FF);
  Spr_Results[BufferIndex++] = (UINT8)(FVC >> 8);
  Spr_Results[BufferIndex++] = (UINT8)(FVC & 0x00FF);
  Spr_Results[BufferIndex++] = (UINT8)(FEV1 >> 8);
  Spr_Results[BufferIndex++] = (UINT8)(FEV1 & 0x00FF);
  
  FEV1_FVC_ratio = (UINT32)(FEV1*100)/FVC;
  
  Spr_Results[BufferIndex++] = (UINT8)(FEV1_FVC_ratio >>8);
  Spr_Results[BufferIndex++] = (UINT8)(FEV1_FVC_ratio & 0x00FF);
  Spr_Results[BufferIndex++] = (UINT8)(PEF >> 8);
  Spr_Results[BufferIndex++] = (UINT8)(PEF & 0x00FF);
  Spr_Results[BufferIndex++] = (UINT8)(FET >> 8);
  Spr_Results[BufferIndex++] = (UINT8)(FET & 0x00FF);
}


static void TimerSampleAdcSpr_Event(void)
{
	UINT16 SprPressureSignalRaw;
	
	// Sample ADC signals
	SprPressureSignalRaw = ADC0_Read12b(ADC_CHANNEL_SPR_PRESSURE_SIGNAL);

	// Perform moving average to avoid noise
	MovingAverage_PushNewValue16b(&SprPressureSignal, SprPressureSignalRaw);
	
	// SprPressureSignal.Result = 	SprPressureSignalRaw; //test if the problem is in moving average
	
	IsnewSampleAvailable = TRUE;
	SwTimer_StartTimer(TimerSampleAdcSpr, SPR_SAMPLING_PERIOD);
}


// Perform the calculation of the spirometry parameters
static void CalculateParameters(void)
{
  static UINT32 Flowfactor = 0;

if (SprActualState == STATE_MEASURING_INHALATION)
  Flow = (SprPressureSignalOffset - SprPressureSignal.Result)*136/10;

else if (SprActualState == STATE_MEASURING_EXHALATION)
  Flow = (SprPressureSignal.Result - SprPressureSignalOffset)*136/10;
	
	if (Flow > 4700)
	  Flowfactor = 0;
	else
	  Flowfactor = (UINT32)((4700-Flow)*100/2700);
	
	Flow = Flow + (Flow*Flowfactor)/100;     // Instantaneous flow [l/s]
	
	if (SprActualState == STATE_MEASURING_INHALATION)
	{
	  FIVC = Volume;
	  if (Flow > PIF) 
	    PIF = Flow;
	}
	
	else if (SprActualState == STATE_MEASURING_EXHALATION)
	{
	  FVC = Volume;
	  if (time_counter_ms == 1000)
	    FEV1 = Volume;
	  if (Flow > PEF)
	    PEF = Flow;
	  FET = time_counter_ms;	  
	}
	
	time_counter_ms += 10;  // Time in milliseconds
  
  Volume = Volume + (Flow/100); // Volume = Volume + (Flow*sampling period) [l]
}


/*******************************
* Public functions
*********************************/

/* call this only once at the beginning of the application */
void Spr_Init(void)
{	
	SprActualState = STATE_INIT;
	
	TimerSampleAdcSpr = SwTimer_CreateTimer(TimerSampleAdcSpr_Event);	
		
	ADC0_Init12b();
	
	IgnoreInitialSamplesCounter = IGNORED_INITIAL_SAMPLES;
	
	// Perform initial measurment to obtain signal offset
	SwTimer_StartTimer(TimerSampleAdcSpr, SENSOR_WARM_UP_TIME);	
}


UINT8 Spr_StartMeasurement(void)
{
	UINT8 status = FALSE;
				 
	if (SprActualState == STATE_IDLE)
	{
		ADC0_Init12b();

                ClearVariables();
		SprActualState = STATE_WAIT_FOR_INHALATION;		
		Spr_IsMeasurementComplete = FALSE;
		Data_type = INH_DATA_TYPE;

		IgnoreInitialSamplesCounter = IGNORED_INITIAL_SAMPLES;
		
		// Start timer to sample ADC 				
		SwTimer_StartTimer(TimerSampleAdcSpr, SENSOR_WARM_UP_TIME);		
		
		status = TRUE;
	}
	else
	{
		status = FALSE;
	}
	return status;
}




void Spr_AbortMeasurement(void)
{
	SwTimer_StopTimer(TimerSampleAdcSpr);
	SprActualState = STATE_IDLE;
        
        AFE_DISABLE();
}

static void ClearVariables(void)
{
  PIF = 0;
  FIVC = 0;
  PEF = 0;
  FVC = 0;
  FEV1 = 0;
  FEV1_FVC_ratio = 0;
  FET = 0;
  Flow = 0;
  Volume = 0;
  time_counter_ms = 0;
  GraphBufferIndex = 0;
}

void Spr_PeriodicTask(void)
{
	/* State machine handler */
	SprStateMachine[SprActualState]();

	/* Event handler */
	if (SprActualEvent != EVENT_SPR_NONE)
	{
		if (Spr_Events[SprActualEvent] != NULL)
		{			
			Spr_Events[SprActualEvent]();	// Execute registered event
			SprActualEvent = EVENT_SPR_NONE;
		}		
	}	
}
